@@ -13,6 +13,7 @@ import sys
 import time
 import weakref
 import asyncio
+import traceback
 
 import glfw
 
@@ -109,6 +110,17 @@ KEY_MAP = {
     glfw.KEY_TAB: "Tab",
 }
 
+KEY_MAP_MOD = {
+    glfw.KEY_LEFT_SHIFT: "Shift",
+    glfw.KEY_RIGHT_SHIFT: "Shift",
+    glfw.KEY_LEFT_CONTROL: "Control",
+    glfw.KEY_RIGHT_CONTROL: "Control",
+    glfw.KEY_LEFT_ALT: "Alt",
+    glfw.KEY_RIGHT_ALT: "Alt",
+    glfw.KEY_LEFT_SUPER: "Meta",
+    glfw.KEY_RIGHT_SUPER: "Meta",
+}
+
 
 class GlfwWgpuCanvas(WgpuCanvasBase):
     """A glfw window providing a wgpu canvas."""
@@ -187,7 +199,7 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
     def _on_close(self, *args):
         all_glfw_canvases.discard(self)
         glfw.hide_window(self._window)
-        self.handle_event({"event_type": "close"})
+        self._emit_event({"event_type": "close"})
 
     def _on_window_dirty(self, *args):
         self._request_draw()
@@ -217,7 +229,7 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
             "height": self._logical_size[1],
             "pixel_ratio": self._pixel_ratio,
         }
-        self.handle_event(ev)
+        self._emit_event(ev)
 
     def _set_logical_size(self, new_logical_size):
         # There is unclarity about the window size in "screen pixels".
@@ -302,6 +314,16 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
     def is_closed(self):
         return glfw.window_should_close(self._window)
 
+    def _emit_event(self, event):
+        try:
+            self.handle_event(event)
+        except Exception:
+            # Print exception and store exc info for postmortem debugging
+            exc_info = list(sys.exc_info())
+            exc_info[2] = exc_info[2].tb_next  # skip *this* function
+            sys.last_type, sys.last_value, sys.last_traceback = exc_info
+            traceback.print_exception(*exc_info)
+
     def handle_event(self, event):
         """Handle an incoming event.
 
@@ -403,7 +425,7 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
             "ntouches": 0,  # glfw dows not have touch support
             "touches": {},
         }
-        self.handle_event(ev)
+        self._emit_event(ev)
 
         self._follow_double_click(action, button)
 
@@ -454,7 +476,7 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
                 "ntouches": 0,  # glfw dows not have touch support
                 "touches": {},
             }
-            self.handle_event(ev)
+            self._emit_event(ev)
 
     def _on_cursor_pos(self, window, x, y):
         # Store pointer position in logical coordinates
@@ -470,7 +492,7 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
             "ntouches": 0,  # glfw dows not have touch support
             "touches": {},
         }
-        self.handle_event(ev)
+        self._emit_event(ev)
 
     def _on_scroll(self, window, dx, dy):
         # wheel is 1 or -1 in glfw, in jupyter_rfb this is ~100
@@ -482,27 +504,20 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
             "y": self._pointer_pos[1],
             "modifiers": list(self._key_modifiers),
         }
-        self.handle_event(ev)
+        self._emit_event(ev)
 
     def _on_key(self, window, key, scancode, action, mods):
 
-        # Map modifier keys, and update self._key_modifiers.
-        # If this callback is for Shift being pressed, then Shift is already in mods.
-        modifiers = []
-        if glfw.MOD_SHIFT & mods:
-            modifiers.append("Shift")
-        if glfw.MOD_CONTROL & mods:
-            modifiers.append("Control")
-        if glfw.MOD_ALT & mods:
-            modifiers.append("Alt")
-        if glfw.MOD_SUPER & mods:
-            modifiers.append("Meta")
-        self._key_modifiers = modifiers
+        modifier = KEY_MAP_MOD.get(key, None)
 
         if action == glfw.PRESS:
             event_type = "key_down"
+            if modifier:
+                self._key_modifiers.add(modifier)
         elif action == glfw.RELEASE:
             event_type = "key_up"
+            if modifier:
+                self._key_modifiers.discard(modifier)
         else:  # glfw.REPEAT
             return
 
@@ -526,7 +541,7 @@ class GlfwWgpuCanvas(WgpuCanvasBase):
             "key": keyname,
             "modifiers": list(self._key_modifiers),
         }
-        self.handle_event(ev)
+        self._emit_event(ev)
 
 
 # Make available under a name that is the same for all gui backends
